--[[
Interface: 5.0.1
Title: LibBalancePowerTracker
Version: 1.1.1
Author: Kurohoshi (EU-Minahonda)

--INFO
	LibBalancePowerTracker is a library designed to provide the foresee energy feature to Balance Druids.
	CPU is only used when required, so the CPU usage is very low (busy waitting forbidden, always on demand).
	
	FORESEE ENERGY:
	Foresee Energy is a feature that analizes the spells you are casting but are yet to be finished and and computes 
	the energy of them. This allows the library to distinguish between two kinds of energy and Eclipse direction: One real, 
	the one you have at the moment and other virtual, the one you'll have when you finish your cast.
	Foresee Energy works assuming the following: 
		-You're going to hit it, if not, it will update as soon as it's known.
		-You're not going to proc Euphoria (2x energy gain).
			If you proc it, it will update immediately, this only means you will reach Eclipse earlier than the library predicted the moment 
			before the Eupforia proc. You can never get an Euphoria proc that pushes you into Eclipse, so Eclipse procs are predicted accurately. 
	All the features with the 'virtual' tag (virtual Energy, virtual Eclipse ...) rely on Foresee Energy.
	
--API
	There are 5 variables related to Eclipse energy used in this addon:
		energy: The energy you have at the moment. Int = [-100,100]
		direction: The direction of the arrow. String = {"none","sun","moon"}
		virtual_energy: The energy you will have when the spell you're casting and all the flying spells land. Int = [-100,100]
		virtual_direction: The direction of the arrow when the spell you're casting and all the flying spells land. String = {"none","sun","moon"}
		virtual_eclipse: -100 if lunar, 100 if solar, false otherwise.

	FUNCTIONS:

	id = LibBalancePowerTracker:RegisterCallback(function(energy, direction, virtual_energy, virtual_direction, virtual_eclipse))
	These callbacks will be fired when there is a change in one of the Eclipse energy variables, usually twice per WR/SF/SS.
	NOTE: When registering a callback, that callback will be fired once.
		
	failed = LibBalancePowerTracker:UnregisterCallback(id)
	Tries to unregister the callback with identifier id (id is returned only when you register the callback).
	
	energy, direction, virtual_energy, virtual_direction, virtual_eclipse = LibBalancePowerTracker:GetEclipseEnergyInfo()
	Gets the current state of the variables.
	
	version,subversion,revision = LibBalancePowerTracker:GetVersion()
	Gets the current working version of the library.	

--Must keep an eye on this:
	AoE silence ----------------------------------- Working in 4.1
	AE buff doesn't reset when changing talents --- True in 5.0.3
	Critters doesn't give energy------------------- True in 5.0.3 (Didn't code check, too much for so less gain)

--CHANGELOG
v 1.1.1 SotF energy prediction while in neutral state fixed
		Fixed interaction between CA and PvP bonus
		Added WoW v check on load
		Added 'select' to locals
		Celestial Aligment & Astral Empowerment check when changing talents
		Added Astral Empowerment mechanics
		Fixed double call on log
		Added debug function
		Fixed AC not erasing as soon as reching 100 energy
		Fixed failed prediction when reaching eclipse and then spending a talent point
		Talent reset/spent correctly handled, even with buffs
		Cancel buffs (AE & CA) support

v 1.1.0 Many changes to update LBPT for WoW 5.0, including new spell mechanics

v 1.0.13 Renamed SpellQueueADT to LBPT_SpellQueueADT

v 1.0.12 Changes in UnitByName(Name)
		 Changed _Sent to account for spell number
		 Fixed not reseting eclipse direction when entering arena (this was fixed in 4.3 by Blizz)
		 PVP 4p bug workaround isn't needed anymore
		 Increased SS and WR autodelete timers to 5s (so it works at max distance on Ragnaros figth)
		 Updated SpellQueueADT-1.1
		 Memory usage no longer increases in combat (Memory leak fixed)
		 Added support to UNIT_DESTROYED event, but commented till it's necessary
		 .toc updated

v 1.0.11 Updated code to handle Euphoria & 4T12 bonus hotfix

v 1.0.10 Simplified some code
		 Sync even when not Balance
		 Checking for tier12 even when not Balance

v 1.0.9 Increased WR and SS autodelete timers by 1s, SF's by .5s
		Fixed advanced settings bugs
		Eclipse stat rewritten (less CPU used)

v 1.0.8 SpellQueueADT updated

v 1.0.7 Tier12 fully supported
		Handling specialWR now here
		Updated SpellQueueADT
		Some code improvements
		Changed reachEnd to virtualEclipse

v 1.0.6	WoW 4.2 fix
		SF autodelete timer increased by .5s
		Updated SpellQueueADT-1.1
		Interrupted spell bug fixed
		Fixed DoTs' energy for 4.2
		Early support for 4t12 bonus
		Included tier set
		Removed some unnecessary functions (RmoveFirsto0...)
		Minor bug in Advanced settings fixed.
		Minor bug causing to fire two callbacks when crossing 0 fixed.

v 1.0.5 Moved UpdateEuphoria into ReCheck
		4.1 fix
		Now shouldnt load when not a druid succeesfully
		Future log compatibility functions
		The mark shouldnt 'dance' at 0,100 and -100 energy anymore.
		Extra functions to avoid letting a spell remain in the queue when it must be erased.
		
v 1.0.4 Changed Euphoria chance based on Hamlet's findings on www.elitistjerks.com 
		Added target of target to the unit check.
		Fixed casting glyph of SS counting as casting SS
		Use spellId instead of names
		
v 1.0.3 Changed to use propperly SpellQueueADT 1.1.2
		Now erases flying spells when teleporting.
		FEATURE: Eclipse chance calculation.
		FEATURE: Energy statistically calculation.

v 1.0.2 Fixed sometimes not fetching the direction properly. 
		Fixed PvP bonus

v 1.0.1 Reduced the number of callbacks fired.

v 1.0.0 Release
--]]

local version = {1,1,1};

if (LibBalancePowerTracker and LibBalancePowerTracker.CompareVersion and LibBalancePowerTracker:CompareVersion(version)) then return; end;

if select(4, GetBuildInfo())<50001 then return; end;

--Initialize Global Lib
LibBalancePowerTracker = {};
function LibBalancePowerTracker:CompareVersion(versionTable) 
	for i,v in ipairs(versionTable) do
		if version[i] < v then
			return false;
		end;
	end;
	return true;
end;

--Locals
----GLOBALS TO LOCALS-------------------------------------------------------------------
local GetEclipseDirection,UnitPower,SPELL_FAILED_NOT_READY,SPELL_FAILED_SPELL_IN_PROGRESS = GetEclipseDirection,UnitPower,SPELL_FAILED_NOT_READY,SPELL_FAILED_SPELL_IN_PROGRESS;
local UnitGUID,UnitBuff,GetTalentInfo,GetSpellInfo = UnitGUID,UnitBuff,GetTalentInfo,GetSpellInfo
local GetInventoryItemID,abs,pairs,ipairs,tonumber,GetSpellCooldown,GetTime,select = GetInventoryItemID,abs,pairs,ipairs,tonumber,GetSpellCooldown,GetTime,select
local LibBalancePowerTracker = LibBalancePowerTracker;
local SPELL_POWER_ECLIPSE,UnitCastingInfo,UnitChannelInfo = SPELL_POWER_ECLIPSE,UnitCastingInfo,UnitChannelInfo;
----DATA--------------------------------------------------------------------------------
------STATIC----------------------------------------------------------------------------
local data ={
	WR  = {name = GetSpellInfo(5176)  , energy = 15, energize_events = 1, spellId=5176  }, -- name & energy Wrath
	SF  = {name = GetSpellInfo(2912)  , energy = 20, energize_events = 1, spellId=2912  }, -- name & energy Starfire
	SS  = {name = GetSpellInfo(78674) , energy = 20, energize_events = 1, spellId=78674 }, -- name & energy StarSurge
	AC	= {name = GetSpellInfo(127663), energy = 15, energize_events = 7, ms_btwn_events = 1000, spellId=127663}, -- name & energy Astral Comunion
	
	EE  			= {spellId = 89265}, -- Eclipse Energy spell
	SuddenEclipse	= {spellId = 95746}, --PvP energy proc
	
	CA				= {name = GetSpellInfo(112071), spellId = 112071}, -- Celestial Aligment buff id
	AE				= {name = GetSpellInfo(130238), spellId = 130238}, -- Astral Empowerment buff id
	LunarEclipse  	= {name = GetSpellInfo(48518) , spellId = 48518 }, -- Lunar eclipse buff id
	SolarEclipse  	= {name = GetSpellInfo(48517) , spellId = 48517 }, -- Solar eclipse buff id
	
	SoulOfTheForest = {talentIndex = 10, bonus = 20},
	balanceTiersItemId ={ --Pieces of bonus that change things
		[12]={
			[1]={ [71108]="n",[71497]="h"},--head
			[3]={ [71111]="n",[71500]="h"},--shoulders
			[5]={ [71110]="n",[71499]="h"},--chest
			[7]={ [71109]="n",[71498]="h"},--trousers
			[10]={[71107]="n",[71496]="h"},--gloves
			bonus4p = true,
			tested4p = false, --Can't test it atm
		},
	}
}
------VARS------------------------------------------------------------------------------
local vars = {
	isDruid = nil,
	isBalance = nil,
	
	spell_casting = 0,
	spell_num = 0,
	energize_events_remaining = 0,
	--next_energize_at = 0,
	last_energize_at = 0,
	
	celestial_lockout_end = 0,
	astral_empowerment_end = 0,
	
	energy = 0,
	vEnergy = 0,
	eclipse = false,
	vEclipse = false,
	direction = "none",
	vDirection = "none",
	
	bug0ToBeWorkarounded = false,

	tiers = {
		[1]=false,--head
		[3]=false,--shoulders
		[5]=false,--chest
		[7]=false,--trousers
		[10]=false,--gloves
		tierPieceCount = {},
	}
}
setmetatable(vars.tiers.tierPieceCount, {__index = function () return 0 end})
local playerGUID,lastCallback,callbacks,number_callbacks = false,0,{},0;
------ENERGY FUNCTIONS------------------------------------------------------------------
--[[If dir == none then
----index is side of bar
--else
----index is eclipse 
--I realize there are some combinations that aren't going to happen]] 
local energyFromSpell={
	[data.SF.spellId]={
		moon	= {
			[-100] = 0,
			[ 000] = 0,
			[ 100] = 0,
		},
		sun		= {
			[-100] = data.SF.energy,
			[ 000] = data.SF.energy,
			[ 100] = data.SF.energy,
		},
		none 	= {
			[-100] = data.SF.energy,
			[ 000] = data.SF.energy,
			[ 100] = data.SF.energy,
		},
	},
	[data.WR.spellId]={
		moon	= {
			[-100] = -data.WR.energy,
			[ 000] = -data.WR.energy,
			[ 100] = -data.WR.energy,
		},
		sun		= {
			[-100] = 0,
			[ 000] = 0,
			[ 100] = 0,
		},
		none 	= {
			[-100] = -data.WR.energy,
			[ 000] = -data.WR.energy,
			[ 100] = -data.WR.energy,
		},
	},
	[data.SS.spellId]={
		moon	= {
			[-100] = -data.SS.energy,
			[ 000] = -data.SS.energy,
			[ 100] = -data.SS.energy,
		},
		sun		= {
			[-100] = data.SS.energy,
			[ 000] = data.SS.energy,
			[ 100] = data.SS.energy,
		},
		none 	= {
			[-100] = -data.SS.energy,
			[ 000] = data.SS.energy,
			[ 100] = data.SS.energy,
		},
	},
	[data.AC.spellId]={
		moon	= {
			[-100] = -data.AC.energy,
			[ 000] = -data.AC.energy,
			[ 100] = -data.AC.energy,
		},
		sun		= {
			[-100] = data.AC.energy,
			[ 000] = data.AC.energy,
			[ 100] = data.AC.energy,
		},
		none 	= {
			[-100] = -data.AC.energy,
			[ 000] = data.AC.energy,
			[ 100] = data.AC.energy,
		},
	},
}
local energizeEventsFromSpell={
	[data.SF.spellId]={
		moon	= 0,
		sun		= data.SF.energize_events,
		none 	= data.SF.energize_events,
	},
	[data.WR.spellId]={
		moon	= data.WR.energize_events,
		sun		= 0,
		none 	= data.WR.energize_events,
	},
	[data.SS.spellId]={
		moon	= data.SS.energize_events,
		sun		= data.SS.energize_events,
		none 	= data.SS.energize_events,
	},
	[data.AC.spellId]={
		moon	= data.AC.energize_events,
		sun		= data.AC.energize_events,
		none 	= data.AC.energize_events,
	},
}
----TIMERS-------------------------------------------------------------------------------
local timers={holder 	= CreateFrame("Frame",nil,UIParent);}
timers.broadcastTier 	= CreateFrame("Cooldown",nil,timers.holder)
timers.delayedUpdate 	= CreateFrame("Cooldown",nil,timers.holder)
----FRAME--------------------------------------------------------------------------------
local frame = CreateFrame("Frame",nil,UIParent); --(loading events & workaround)
local combatFrame = CreateFrame("Frame",nil,UIParent); --Combat events
local LBPT = {};
-----------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------
--Aux functions
local spellsUsed = {
	[data.WR.name] = tonumber(data.WR.spellId),
	[data.SS.name] = tonumber(data.SS.spellId),
	[data.SF.name] = tonumber(data.SF.spellId),
	[data.AC.name] = tonumber(data.AC.spellId),
	[data.WR.spellId] = tostring(data.WR.name),
	[data.SS.spellId] = tostring(data.SS.name),
	[data.SF.spellId] = tostring(data.SF.name),
	[data.AC.spellId] = tostring(data.AC.name),
}
local function deleteSpell()
	vars.spell_casting = 0;
	vars.spell_num = 0;
	vars.energize_events_remaining = 0;
	LBPT.RecalcEnergy()
end
local function CheckEcplipseBuff() 
	vars.eclipse = (UnitBuff('player',data.SolarEclipse.name) and 100) or (UnitBuff('player',data.LunarEclipse.name) and -100)
	return vars.eclipse
end
local function checkCelestialAligmentBuff()
	local tim_end = select(7,UnitBuff('player',data.CA.name)) or 0; 
	vars.celestial_lockout_end = tim_end*1000;
	return vars.celestial_lockout_end;
end
local function checkAstralEmpowermentBuff()
	local tim_end = select(7,UnitBuff('player',data.AE.name)) or 0;
	vars.astral_empowerment_end = tim_end*1000;
	return vars.astral_empowerment_end;
end
local function astralEmpowermentBonus(spellEnergy,spellId) --Astral Empowerment bonus, returns energy with bonus added
	if spellEnergy == 0 then
		return 0;
	end
	if spellId == data.SF.spellId then
		return spellEnergy + 7;
	elseif spellId == data.WR.spellId then
		return spellEnergy - 5;
	elseif spellId == data.SS.spellId then
		if spellEnergy > 0 then
			return spellEnergy + 7
		else
			return spellEnergy - 7
		end
	else
		return spellEnergy;
	end
end
local function UpdateEnergizeAt(channeled)
	if channeled then
		vars.last_energize_at = select(6,UnitChannelInfo("player"))
		--vars.next_energize_at = vars.last_energize_at - data.AC.ms_btwn_events * (vars.energize_events_remaining-1)
	else
		vars.last_energize_at = select(6,UnitCastingInfo("player"))
		--vars.next_energize_at = vars.last_energize_at
	end
end

do --Loading
	frame:SetScript("OnEvent",  function(_, event, ...) LBPT[event](...) end);
	frame:RegisterEvent("PLAYER_LOGIN");
	LBPT.postLogin = false;
	
	function LBPT.PLAYER_LOGIN()
		LBPT.postLogin = true;		
		vars.isDruid = select(2,UnitClass("player"))=="DRUID"; 
		if vars.isDruid and number_callbacks ~= 0 then 
			LBPT.load()
		end 
	end;
	
	function LBPT.load()
		--on reload, info working ok
		--on load, no
		playerGUID=UnitGUID("player");
		
		frame:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED"); 
			
		--Check tier
		frame:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
		for k in pairs(vars.tiers) do if tonumber(k) then LBPT.PLAYER_EQUIPMENT_CHANGED(k,GetInventoryItemID("player", k)) end end
		
		vars.isBalance = (GetSpellCooldown(data.SS.name)~=nil)
		LBPT.RegisterCombatEvents(vars.isBalance)
		LBPT.Reset(vars.isBalance)
					
		if vars.isBalance and (not vars.eclipse) and vars.direction == "none" then --balance, Login & no conseguimos los valores actualizados
			frame:RegisterEvent("ECLIPSE_DIRECTION_CHANGE"); --not fired when /reload, used to check eclipse & dir on load
			timers.delayedUpdate:SetScript("OnHide",function() frame:UnregisterEvent("ECLIPSE_DIRECTION_CHANGE"); end)
			timers.delayedUpdate:SetCooldown(GetTime(),0) 
		end
	end
	
	function LBPT.unload()
		frame:UnregisterAllEvents();
		combatFrame:UnregisterAllEvents();
	end
	
	function LBPT.ECLIPSE_DIRECTION_CHANGE()
		LBPT.Reset(true)
		frame:UnregisterEvent("ECLIPSE_DIRECTION_CHANGE");
	end; 

	function LBPT.RegisterCombatEvents(balanceNow)
		if balanceNow then
			for k in pairs(LBPT.combat) do
				combatFrame:RegisterEvent(k);
			end
		else 
			for k in pairs(LBPT.combat) do
				combatFrame:UnregisterEvent(k);
			end
		end
	end
	
	function LBPT.Reset(balanceNow)
		--Reset values
		vars.energy = 0
		vars.vEnergy = 0
		vars.eclipse = false
		vars.vEclipse = false
		vars.direction = "none"
		vars.vDirection = "none"
		vars.bug0ToBeWorkarounded = false;
		vars.celestial_lockout_end = 0;
		vars.astral_empowerment_end = 0;
		
		--Propagate values
		if balanceNow then 	
			vars.direction = GetEclipseDirection() or vars.direction --Only useful after login, not accurate when changing talents
			CheckEcplipseBuff() --Only useful after login	
			checkCelestialAligmentBuff()
			checkAstralEmpowermentBuff()
			LBPT.RecalcEnergy(); --energy is not accurate when changing talens --Need to avoid 0bug
		else
			LBPT.FireCallbacks()
		end
	end
end

do --Talent check/change events
	function LBPT.PLAYER_SPECIALIZATION_CHANGED()
		local balanceNow = GetSpellCooldown(data.SS.name)~=nil;
		
		if vars.isBalance and balanceNow and GetEclipseDirection() then 
			--GetEclipseDirection is nil o teleport, handled by its own event, put here to avoid firing this event when portaling
			LBPT.Reset(balanceNow)
			
			if  UnitPower("player",SPELL_POWER_ECLIPSE) ~= 0  then
				frame:RegisterEvent("UNIT_POWER"); --Si al pasar de pollo a pollo tenemos energia, retrasar un evento para actualizar cuando vuelva a 0
			end
		end
		
		if vars.isBalance ~= balanceNow then --Si cambiamos de pollo a chopo o viceversa, hemos de actualizar los eventos y resetear los valores
			vars.isBalance = balanceNow
			LBPT.RegisterCombatEvents(balanceNow)
			LBPT.Reset(balanceNow)
		end
	end
	
	function LBPT.UNIT_POWER(unit,power)
		if unit == "player" and power == "ECLIPSE" then
			frame:UnregisterEvent("UNIT_POWER");
			vars.bug0ToBeWorkarounded = false;
			local e =  UnitPower("player",SPELL_POWER_ECLIPSE)
			if vars.energy ~= e then
				LBPT.RecalcEnergy(e) --Need to avoid 0bug
			end
		end
	end
end

do --Combat events-------------
	LBPT.combat = {};
	
	do --Combat frame script creation
		local combatTable = LBPT.combat;
		combatFrame:SetScript("OnEvent",  function(_, event, ...) combatTable[event](...) end);
	end
	
	local function startSpell(unit,num,id,channeled)
		if unit == "player" and spellsUsed[id] then
			vars.spell_casting = id;
			vars.spell_num = num;
			vars.energize_events_remaining = energizeEventsFromSpell[id][vars.direction];
				
			UpdateEnergizeAt(channeled)
			
			if vars.bug0ToBeWorkarounded then --need to avoid 0bug 
				LBPT.RecalcEnergy(vars.energy) 
			else
				LBPT.RecalcEnergy()
			end
		end
	end
	
	function LBPT.combat.UNIT_SPELLCAST_START(unit,_,_,num,id)			startSpell(unit,num,id,false)	end
	function LBPT.combat.UNIT_SPELLCAST_CHANNEL_START(unit,_,_,num,id)	startSpell(unit,num,id,true )	end
	
	function LBPT.combat.UNIT_SPELLCAST_DELAYED(unit,_,_,num,id)
		if unit == "player" and spellsUsed[id] and vars.celestial_lockout_end ~= 0 and vars.spell_num == num and vars.spell_casting == id then
			UpdateEnergizeAt(false)
			LBPT.RecalcEnergy()
		end
	end
	
	function LBPT.combat.UNIT_SPELLCAST_CHANNEL_UPDATE(unit,_,_,num,id)
		if unit == "player" and spellsUsed[id] and vars.celestial_lockout_end ~= 0 and vars.spell_num == num and vars.spell_casting == id then
			UpdateEnergizeAt(true)
			LBPT.RecalcEnergy()
		end
	end
	
	local function finishSpell(unit,num,id)
		if unit == "player" and spellsUsed[id] and vars.spell_num == num and vars.spell_casting == id then
			deleteSpell()
		end
	end
		
	function LBPT.combat.UNIT_SPELLCAST_SUCCEEDED(unit,_,_,num,id)		finishSpell(unit,num,id) end
	function LBPT.combat.UNIT_SPELLCAST_CHANNEL_STOP(unit,_,_,num,id)	finishSpell(unit,num,id) end

	function LBPT.combat.PLAYER_DEAD() --Reset queue & clear energy upon death
		vars.direction = GetEclipseDirection() or "none"; 
		deleteSpell()
	end
	
	do --COMBAT LOG HANDLER -------------------------------------------------------		
		local auraLossFunctionTable = {
			[data.LunarEclipse.spellId] = 	function() if vars.eclipse then vars.eclipse = false LBPT.RecalcEnergy() end end,
			[data.SolarEclipse.spellId] = 	function() if vars.eclipse then vars.eclipse = false LBPT.RecalcEnergy() end end,
			[data.CA.spellId] 			= 	function() if vars.celestial_lockout_end ~=0 then vars.celestial_lockout_end = 0; LBPT.RecalcEnergy() end end,
			[data.AE.spellId] 			= 	function() if vars.astral_empowerment_end ~= 0 then vars.astral_empowerment_end = 0; LBPT.RecalcEnergy() end end,
		}
		
		local function UpdateACTicksLeft()
			vars.energize_events_remaining = math.floor(vars.last_energize_at/1000 - GetTime()+.5)
			--print(vars.last_energize_at/1000,GetTime(),vars.energize_events_remaining,"ticks remaining")
			return vars.energize_events_remaining
		end
		
		local unfilteredCombatLogTable = {
			SPELL_ENERGIZE 		= function(id,amount,typeEnergy)	
									if (typeEnergy == SPELL_POWER_ECLIPSE) then
										local energy = UnitPower("player" , SPELL_POWER_ECLIPSE);
										
										if (vars.energy == energy) then
											--If there's no real energy gain, there is no update
											return
										end
										
										if id == data.EE.spellId then
											if vars.celestial_lockout_end == 0 then 
												--SOTF interaction with CA, 
												--to not to erase spells when the energy is CA's
												
												vars.energize_events_remaining = vars.energize_events_remaining -1;
												if vars.energize_events_remaining <=0 or (UpdateACTicksLeft() <= 0) or (abs(energy)==100) then
													--Updating ticks on AC using time due to not having energize events during CA
													--A lot of things (abs(energy)...) rely on the fact that AC is the only spell with more than 1 energize event
													
													vars.spell_casting = 0;
													vars.spell_num = 0;
													vars.energize_events_remaining = 0;
												--else
													--UpdateEnergizeAt(true)
												end
											else
												--Only remove CA if gained by this event, since CA only prevents gaining energy from this spell
												vars.celestial_lockout_end = 0 --CA no more
											end
										end
										
										if energy ==100 then 
											vars.direction = "moon"; 		
											vars.eclipse = 100
										elseif energy == -100 then 
											vars.direction = "sun";	
											vars.eclipse = -100
										elseif (vars.direction == "moon" and energy <=0) or (vars.direction == "sun"  and energy >=0) or (vars.direction == "none") then
											vars.eclipse=false; 
										end
										
										if (energy == 0 and vars.energy+amount ~= 0) or vars.bug0ToBeWorkarounded then --need to check 0bug (and SS)
											vars.bug0ToBeWorkarounded = true;
											frame:RegisterEvent("UNIT_POWER"); --Schedule energy check
											LBPT.RecalcEnergy(vars.energy+amount)  --predict with the energy we have
										else
											LBPT.RecalcEnergy(energy) 
										end
									end
								end,
			SPELL_CAST_FAILED 	= function(id,msg) 
									if vars.spell_casting == id and msg ~= SPELL_FAILED_NOT_READY and msg ~= SPELL_FAILED_SPELL_IN_PROGRESS then 
										deleteSpell()
									end 
								end,
			SPELL_AURA_APPLIED 	= function(id)
									if id == data.AE.spellId then
										checkAstralEmpowermentBuff()
										--There's no need to Recalc energy bc you can't gain this while casting
									elseif id == data.CA.spellId then
										checkCelestialAligmentBuff()
										
										if UnitPower("player",SPELL_POWER_ECLIPSE) ~= 0 then
											frame:RegisterEvent("UNIT_POWER");
										end
										LBPT.RecalcEnergy(0) 
									end
								end,
			SPELL_AURA_REMOVED 	= function(id)
									local f = auraLossFunctionTable[id]
									if f then f(); end
								end,
		}
		function LBPT.combat.COMBAT_LOG_EVENT_UNFILTERED(_,event,_,gUIDor,_,_,_,destGUID,_,_,_,spellId,_,_,amountEnergy,typeEnergy)
			if (gUIDor == playerGUID) then 
				event = unfilteredCombatLogTable[event] 
				if event then event(spellId,amountEnergy,typeEnergy) end
			elseif (destGUID == playerGUID) and event=="SPELL_INTERRUPT" and vars.spell_casting == amountEnergy then 
				deleteSpell()
			end
		end
	end

	do --Direction & energy when teleporting  
		function LBPT.combat.PLAYER_ENTERING_WORLD()
			vars.direction = GetEclipseDirection() or "none";
			deleteSpell()
		end; 
	end
end

do --Recalc Energy function
	local function isActiveAstralEmpowermentAt(timestamp)
		--Checking refresh here, can't in combat log bc applied dose for buffs doesn't fire, 
		--and since I moved this here, I also included refresh check here
		if vars.astral_empowerment_end == 0 then
			return false;
		end
		checkAstralEmpowermentBuff()
		return vars.astral_empowerment_end >= timestamp;
	end
	
	local SotFIndex = data.SoulOfTheForest.talentIndex;
	local SotFBonus = data.SoulOfTheForest.bonus;
	local function ExtraEnergy(energy,direction) --computes extra energy
		local eclipse = vars.eclipse
		local newEnergy;
		
		--Get spell energy
		if vars.energize_events_remaining > 0 and vars.last_energize_at >= vars.celestial_lockout_end then
			local spellEnergy = energyFromSpell[vars.spell_casting][direction][((direction == "none") and ((energy>=0 and 100) or -100)) or (eclipse or 0) ];

			if isActiveAstralEmpowermentAt(vars.last_energize_at) then
				newEnergy = energy + astralEmpowermentBonus(spellEnergy,vars.spell_casting);
			else
				newEnergy = energy + spellEnergy;
			end
		else
			newEnergy = energy;
		end
		
		--Add SOTF energy
		if select(5,GetTalentInfo(SotFIndex)) then --SOTF
			if vars.celestial_lockout_end ~= 0 then --CA 
				if direction == "moon" then
					newEnergy = newEnergy -SotFBonus;
				else
					newEnergy = newEnergy +SotFBonus;
				end
			elseif energy > 0 and newEnergy <=0 then
				newEnergy = newEnergy -SotFBonus;
			elseif energy < 0 and newEnergy >=0 then
				newEnergy = newEnergy +SotFBonus;
			end
		end
		
		--Set energy boundaries
		local newDirection;
		if newEnergy>=100 then 
			newEnergy,newDirection,eclipse = 100,"moon",100;
		elseif newEnergy<=-100 then 
			newEnergy,newDirection,eclipse = -100,"sun",-100;
		elseif (direction == "moon" and newEnergy <=0) or (direction == "sun"  and newEnergy >=0) or (direction == "none") then
			eclipse = false;
			newDirection = direction
		else
			newDirection = direction
		end

		
		if energy ~= newEnergy then 
			--if there's no change in energy, prediction shouln't change eclipse or direction
			--It should be here, because if you can only get 0 energy from spell, due to being already
			--at max energy, you won't get eclipse.
			--Test by reaching eclipse and then spend a "talent point"
			--CPU effect: 
				-- +1 asingacin
				-- +2 asignacin /ciclo
				-- +1 comprobacin
			---- total: +2 instrucciones: incremento irrelevante
			return newEnergy,newDirection,eclipse;
		else
			return energy,direction,vars.eclipse;
		end
	end
	
	function LBPT.RecalcEnergy(energy)
		if not energy then  energy = UnitPower("player" , SPELL_POWER_ECLIPSE); end
		vars.energy = energy;
		vars.vEnergy,vars.vDirection,vars.vEclipse = ExtraEnergy(energy,vars.direction)	
		LBPT.FireCallbacks()
	end
end

do--Calling callbacks functions
	function LBPT.FireCallbacks()
		for k,v in pairs(callbacks) 		
			do v(vars.energy,vars.direction,vars.vEnergy,vars.vDirection,vars.vEclipse); 	
		end; 
	end;
end

do--Called functions (API)
	function LibBalancePowerTracker:RegisterCallback(callback)
		if number_callbacks==0 and LBPT.postLogin and vars.isDruid then
			LBPT.load();
		end
		
		number_callbacks=number_callbacks+1

		lastCallback=lastCallback+1
		callbacks[lastCallback]=callback;

		callback(vars.energy,vars.direction,vars.vEnergy,vars.vDirection,vars.vEclipse)
		return lastCallback
	end
	function LibBalancePowerTracker:UnregisterCallback(id)
		if callbacks[id] then
			callbacks[id]=nil;
		else
			return true;
		end
		
		number_callbacks=number_callbacks-1
		
		if number_callbacks==0 then
			LBPT.unload();
		end
	end
	function LibBalancePowerTracker:GetVersion()	return version[1],version[2],version[3]; end;
	
	do --GetEclipseEnergyInfo
		--If called once, it would be called again, so I create a callback inside this do-end block
		local e,d,vE,vD,vEc;
		local callbackId = nil
		
		function LibBalancePowerTracker:GetEclipseEnergyInfo()
			if not callbackId then
				callbackId = self:RegisterCallback(
					function(energy,direction,vEnergy,vDirection,vEclipse) 
						e,d,vE,vD,vEc = energy,direction,vEnergy,vDirection,vEclipse 
					end 
				)
			end
		
			return e,d,vE,vD,vEc
		end
	end
end

----TIER MODIFIER FUCNTION (At the end, so it sees and can modify all locals, sometimes I'll need to remove do-end blocks)
do --Tier bonus check
	local broadcasted = {}
	setmetatable(broadcasted, {__index = function () return 0 end})
	timers.broadcastTier:SetScript("OnHide",function() 
		for k,v in pairs(vars.tiers.tierPieceCount) do
			if broadcasted[k] ~= v then --broadcast
				if broadcasted[k] < v then --gained bonus
					for i = broadcasted[k]+1,v do
						if data.balanceTiersItemId[k]["bonus"..i.."p"] then 
							if data.balanceTiersItemId[k]["tested"..i.."p"] then
								print("|c00a080ffLibBalancePowerTracker|r: Tier"..k.." "..i.."p bonus detected.") 
							else
								print("|c00a080ffLibBalancePowerTracker|r: Tier"..k.." "..i.."p bonus detected (Untested).") 
							end
						end
					end
				else --lost bonus
					for i = v+1,broadcasted[k] do
						if data.balanceTiersItemId[k]["bonus"..i.."p"] then 
							print("|c00a080ffLibBalancePowerTracker|r: No tier"..k.." "..i.."p bonus detected.") 
						end
					end
				end
				broadcasted[k] = v
			end
		end
	end)
	
	function LBPT.PLAYER_EQUIPMENT_CHANGED(slot,hasItem)
		local setInSlot = vars.tiers[slot]
		if setInSlot then						--print("retirado objeto de "..slot)
			vars.tiers[slot]=false;
			
			local num_p_tier = vars.tiers.tierPieceCount[setInSlot]-1
			vars.tiers.tierPieceCount[setInSlot]= num_p_tier --print("Tienes "..vars.tiers.tierPieceCount[setInSlot].." piezas de tier "..setInSlot)
			
			if data.balanceTiersItemId[setInSlot]["bonus"..(num_p_tier+1).."p"] and LBPT.BonusTier[setInSlot][num_p_tier+1].Off() then 
				timers.broadcastTier:SetCooldown(GetTime(),0)  
			end	
		end
		
		if hasItem and setInSlot ~= nil then 				--print("se intenta poner una pieza en "..slot)
			local id = GetInventoryItemID("player", slot)	--print("el id de la pieza es "..tostring(id))
			for k,v in pairs(data.balanceTiersItemId) do	--print("buscando en tier "..k)
				if v[slot] and v[slot][id] then				--print("encontrado en tier "..k)
					vars.tiers[slot]=k;
					
					local num_p_tier = vars.tiers.tierPieceCount[k]+1
					vars.tiers.tierPieceCount[k]= num_p_tier --print("Tienes "..vars.tiers.tierPieceCount[k].." piezas de tier "..k);
					
					if v["bonus"..num_p_tier.."p"] and LBPT.BonusTier[k][num_p_tier].On() then 
						timers.broadcastTier:SetCooldown(GetTime(),0)
					end
					
					return
				end
			end
		end
	end
end

LBPT.BonusTier={
	[12]={
		[4]={
			On  = 	function()
						energyFromSpell[data.SF.spellId].sun[100]  = data.SF.energy+5
						energyFromSpell[data.SF.spellId].sun[000]  = data.SF.energy+5
						
						energyFromSpell[data.SF.spellId].none[-100] = data.SF.energy+5
						energyFromSpell[data.SF.spellId].none[ 000] = data.SF.energy+5
						energyFromSpell[data.SF.spellId].none[ 100] = data.SF.energy+5

						energyFromSpell[data.WR.spellId].moon[-100] = -data.WR.energy-3
						energyFromSpell[data.WR.spellId].moon[ 000] = -data.WR.energy-3
						
						energyFromSpell[data.WR.spellId].none[-100] = -data.WR.energy-3
						energyFromSpell[data.WR.spellId].none[ 000] = -data.WR.energy-3
						energyFromSpell[data.WR.spellId].none[ 100] = -data.WR.energy-3

						return true;
					end,
			Off = 	function() 
						energyFromSpell[data.SF.spellId].sun[100]  = data.SF.energy
						energyFromSpell[data.SF.spellId].sun[000]  = data.SF.energy
						
						energyFromSpell[data.SF.spellId].none[-100] = data.SF.energy
						energyFromSpell[data.SF.spellId].none[ 000] = data.SF.energy
						energyFromSpell[data.SF.spellId].none[ 100] = data.SF.energy

						energyFromSpell[data.WR.spellId].moon[-100] = -data.WR.energy
						energyFromSpell[data.WR.spellId].moon[ 000] = -data.WR.energy
						
						energyFromSpell[data.WR.spellId].none[-100] = -data.WR.energy
						energyFromSpell[data.WR.spellId].none[ 000] = -data.WR.energy
						energyFromSpell[data.WR.spellId].none[ 100] = -data.WR.energy
						
						return true;
					end,
		},
	},
}


do ----DEBUG---------------------
	local debug_on_load = false;
	------------------------------
	local dbug = false;
	local dbug_init = false;
	local ev;
	function LibBalancePowerTracker:ToogleDebug()
		dbug = not dbug;
		print("|c00a080ffBalancePowerTracker|r: Debug: ",dbug)

		if not dbug_init then
			function LBPT.FireCallbacks()
				if dbug then
					print(vars.energy,vars.direction,vars.vEnergy,vars.vDirection,vars.vEclipse,ev) 
				end
				for k,v in pairs(callbacks) 		
					do v(vars.energy,vars.direction,vars.vEnergy,vars.vDirection,vars.vEclipse); 	
				end; 
			end;
			
			combatFrame:SetScript("OnEvent",  function(_, event, ...) ev=event LBPT.combat[event](...)  end);
			frame:SetScript("OnEvent",  function(_, event, ...) ev=event LBPT[event](...) end);
		end
		dbug_init = true;
	end
	if debug_on_load then
		LibBalancePowerTracker:ToogleDebug()
	end
end




